import logging
import os
import json

from docopt import ParsedOptions

from cli.config_constants import CONFIG_TEMPLATE, CLI_INPUTS_TEMPLATE
from cli.constants import CONFIG_PATH, SUPPORTED_COMMANDS
from cli.services.validators import is_config_file_valid, is_cluster_id


class ConfigObject:
    def __init__(self, parsed_inputs: ParsedOptions, config_file_path=CONFIG_PATH):
        self.config_file_path = config_file_path
        self._logger = logging.getLogger(__name__)
        self.app_config = None
        self._setup_config()
        self.app_inputs = CLI_INPUTS_TEMPLATE
        self._populate_parsed_inputs_data(parsed_inputs)
        self.custom_cid = None

    def create_empty_template(self):
        self._logger.info(f"Creating empty {CONFIG_PATH} file")
        with open(self.config_file_path, "w") as config_file:
            json.dump(CONFIG_TEMPLATE, config_file)

    def _populate_parsed_inputs_data(self, parsed_inputs: ParsedOptions):
        self.app_inputs["demo"] = parsed_inputs["demo"]
        self.app_inputs["demo_subcommand"] = parsed_inputs["<on | off | refresh>"]
        self.app_inputs["cluster_id"] = parsed_inputs["--cluster_id"]
        self.app_inputs["help"] = parsed_inputs["--help"]
        self.app_inputs["debug"] = parsed_inputs["--debug"]

        if self.app_inputs["cluster_id"]:
            if is_cluster_id(self.app_inputs["cluster_id"]):
                self._logger.info(f"Custom cluster id={self.app_inputs['cluster_id']} will be used")
                self.custom_cid = self.app_inputs["cluster_id"]
            else:
                raise RuntimeError("--cluster_id did not match expected format")

        for command, value in parsed_inputs.items():
            if command in SUPPORTED_COMMANDS and value:
                self.app_inputs["command"] = command

    def _setup_config(self):
        if not os.path.exists(self.config_file_path):
            self.create_empty_template()
            self._logger.critical("config.json was missing. Created an empty template.")
            raise RuntimeError("config.json was missing. Created an empty template.")

        with open(self.config_file_path, "r") as config_file:
            configuration_data = json.load(config_file)
            if not is_config_file_valid(configuration_data):
                self._logger.critical("Invalid configuration file structure")
                raise RuntimeError("Invalid configuration file structure")
            self.app_config = configuration_data



"""
se-cli: The swiss army command-line for Cast.AI SE`s.

Usage:
    se-cli demo <on|off|refresh> [--cluster_id <cluster_id>] [-h|--help] [-d|--debug]
    se-cli snapshot <analyze> [--cluster_id <cluster_id>] [-h|--help] [-d|--debug]

Options:
    -h, --help  Show this help message and exit.
    -d, --debug  Enable debug logging.
    -c, --cluster_id <cluster_id>  (Optional) Specify the cluster ID for the demo environment.

Commands:
    demo      Manage the demo environment.
    snapshot  Manage snapshots.

Subcommands for "demo":
    on   Prep demo environment for demo.
    off  Hibernate demo environment.
"""
import sys

from orchestrators.demo_orch import DemoOrchestrator
from orchestrators.snapshot_orch import SnapshotOrchestrator
from services.misc_svc import init

from docopt import docopt


commands_table = {
    "demo": DemoOrchestrator,
    "snapshot": SnapshotOrchestrator
}

if __name__ == '__main__':
    try:
        # TODO: check version once pypi is up
        parsed_options = docopt(__doc__, sys.argv[1:])
        cfg = init(parsed_options)
        orchestrator_class = commands_table[cfg.app_inputs["command"]]
        main_orch = orchestrator_class(cfg)
        main_orch.execute()
    except Exception as e:
        print(f"An error occurred: {e}")



import logging
import time

from cli.models.config import ConfigObject
from cli.orchestrators.base_orch import BaseOrchestrator
from cli.services.validators import validate_kubectl_context

from cast_ai.se.contollers.castai_controller_svc import CastController
from cast_ai.se.contollers.eks_controller_svc import EKSController
from cast_ai.se.contollers.kubectl_controller_svc import KubectlController
from cast_ai.se.models.execution_status import ExecutionStatus
from cast_ai.se.models.cloud_confs import AwsConfig


class DemoOrchestrator(BaseOrchestrator):
    def __init__(self, cfg: ConfigObject):
        super().__init__(cfg)
        self._logger = logging.getLogger(__name__)

        cid = cfg.custom_cid if cfg.custom_cid else cfg.app_config["CAST"]["DEFAULT_CLUSTER_ID"]

        self._cast_ctrl = CastController(cfg.app_config["CAST"]["CAST_API_KEY"], cid)
        self._kubectl_ctrl = KubectlController()

        self._initial_nodes = self._cast_ctrl.get_nodes()

        self._cloud_ctrl = None
        self._set_cloud_controller()
        self.demo_subcommand_mapping = {
            "on": self.demo_on_sequence,
            "off": self.demo_off_sequence,
            "refresh": self.demo_refresh_sequence,
        }
        validate_kubectl_context(self._kubectl_ctrl, self._cast_ctrl.cluster['providerType'])

    def execute(self) -> None:
        subcommand = self._cfg.app_inputs["demo_subcommand"]
        if subcommand in self.demo_subcommand_mapping:
            self.demo_subcommand_mapping[subcommand]()
        else:
            raise ValueError(f'Invalid option: {subcommand}')

    def demo_refresh_sequence(self):
        if self._cast_ctrl.are_policies_disabled():
            self._logger.critical("Policies are off - Cluster hibernating")
            raise RuntimeError("Policies are off - Cluster hibernating (Consider [on] subcommand)")

        self._logger.info(f"{'=' * 80}< Starting demo-REFRESH sequence >{'=' * 80} ")
        start_time = time.time()
        demo_node_count = self._cfg.app_config["GENERAL"]["DEMO_NODE_COUNT"]
        demo_replica_count = self._cfg.app_config["GENERAL"]["DEMO_REPLICAS"]

        initial_nodes = self._cast_ctrl.get_nodes()
        demo_node_count = demo_node_count + len(initial_nodes["items"])

        if not self._cast_ctrl.is_downscaler_disabled():
            self.spinner_run("Disabling downscaler policy", lambda: self._cast_ctrl.disable_downscaler_policy())

        self.spinner_run(f'Scaling cloud default ng to {demo_node_count}',
                         lambda: self._cloud_ctrl.scale_default_ng(demo_node_count))

        all_demo_nodes_up = self.spinner_run("Waiting for new nodes to be ready",
                                             lambda: self._wait_for_new_nodes(len(initial_nodes["items"]),
                                                                              demo_node_count))

        if all_demo_nodes_up.success:
            self.spinner_run(f'Deleting nodes prior to refresh ({len(initial_nodes["items"])})',
                             lambda: self._cast_ctrl.delete_nodes(initial_nodes))
        else:
            print("❗❕ You may want to delete the nodes that existed prior to refresh")

        print(f"It took {int(time.time() - start_time)} seconds to refresh demo")

    def demo_off_sequence(self):
        if self._cast_ctrl.are_policies_disabled():
            self._logger.critical("Policies are off - Cluster hibernating")
            raise RuntimeError("Policies are off - Cluster hibernating")
        demo_off_cronjob = self._cfg.app_config["KUBECTL"]["DEMO_OFF_CRONJOB"]
        self._logger.info(f"{'=' * 80}< Starting demo-OFF sequence >{'=' * 80} ")

        msg = "Scaling all deployments in default namespace to 0"
        self.spinner_run(msg, lambda: self._kubectl_ctrl.scale_deployments(0))

        self.spinner_run("Disabling CAST Unscheduled pods policy",
                         lambda: self._cast_ctrl.disable_unscheduled_pods_policy())

        msg = f"Manually triggering {demo_off_cronjob}"
        cronjob_triggered = (
            self.spinner_run(msg, lambda: self._kubectl_ctrl.trigger_cronjob(demo_off_cronjob, exec_wait=True)))

        if cronjob_triggered.success:
            self.spinner_run("Turning off autoscaling on default ng", lambda: self._cloud_ctrl.scale_default_ng(0))
        else:
            print("❗❕ You may want to turn off autoscaling manually...")

    def demo_on_sequence(self):
        if not self._cast_ctrl.are_policies_disabled():
            self._logger.critical("Policies are on - Cluster not properly Hibernated")
            raise RuntimeError("Policies are on - Cluster not properly Hibernated (Consider [refresh] subcommand)")

        self._logger.info(f"{'=' * 80}< Starting demo-ON sequence >{'=' * 80}")
        start_time = time.time()
        demo_node_count = self._cfg.app_config["GENERAL"]["DEMO_NODE_COUNT"]
        demo_replicas = self._cfg.app_config["GENERAL"]["DEMO_REPLICAS"]
        initial_nodes = self._cast_ctrl.get_nodes()

        self.spinner_run(f"Scaling cloud default ng to {demo_node_count}",
                         lambda: self._cloud_ctrl.scale_default_ng(demo_node_count))
        all_demo_nodes_up = (self.spinner_run("Waiting for new nodes to be ready",
                             lambda: self._wait_for_new_nodes(len(initial_nodes["items"]), demo_node_count)))
        self.spinner_run("Disabling CAST Binpacking policies",
                         lambda: self._cast_ctrl.disable_downscaler_policy())
        self.spinner_run("Enabling CAST policies",
                         lambda: self._cast_ctrl.enable_existing_policies())
        self.spinner_run(f"Scaling all deployments to {demo_replicas} replicas",
                         lambda: self._kubectl_ctrl.scale_deployments(demo_replicas))
        if all_demo_nodes_up.success:
            self.spinner_run(f'Deleting initial {len(initial_nodes["items"])} fallback nodes ',
                             lambda: self._cast_ctrl.delete_nodes(initial_nodes))
        else:
            print("❗❕ You may want to delete the Fallback nodes...")

        print(f"It took {int(time.time() - start_time)} seconds to prep demo")

    def _wait_for_new_nodes(self, initial_node_count: int, new_demo_count: int) -> ExecutionStatus:
        try:
            self._logger.info(f"{'-' * 70}[ Waiting for new nodes to be ready ]")
            # demo_node_count = self._cfg.app_config["GENERAL"]["DEMO_NODE_COUNT"]
            ng_scaling_timeout = self._cfg.app_config["GENERAL"]["NG_SCALING_TIMEOUT"]
            start_time = time.time()
            while True:
                nodes = self._cast_ctrl.get_nodes()
                self._logger.debug(f"Current:{len(nodes['items'])} "
                                   f"Initial:{initial_node_count} "
                                   f"Requested:{new_demo_count}")
                if len(nodes["items"]) == initial_node_count + new_demo_count:
                    all_nodes_ready = all(node["state"]["phase"] == "ready" for node in nodes["items"])
                    if all_nodes_ready:
                        self._logger.info(f"Waited {time.time() - start_time} for all new nodes to be ready...")
                        return ExecutionStatus()
                    else:
                        node_states = [node["state"]["phase"] for node in nodes["items"]]
                        node_states_str = ", ".join(node_states)
                        self._logger.debug(f"Node states: ({node_states_str})")
                if time.time() - start_time > ng_scaling_timeout:
                    self._logger.warning(f"Timeout ({ng_scaling_timeout}sec) reached while waiting for new nodes")
                    return ExecutionStatus(f"Timeout ({ng_scaling_timeout}sec)")
                time.sleep(1)
        except Exception as e:
            self._logger.error(f"Something went wrong:[{str(e)}]")
            return ExecutionStatus(f"Something went wrong:[{str(e)}]")

    def _set_cloud_controller(self):
        provider_type = self._cast_ctrl.cluster['providerType']
        #match (self._cast_ctrl.cluster['providerType']):
        if provider_type == "eks":
            self._cloud_ctrl = EKSController(AwsConfig(self._cfg.app_config["AWS"]))
            self._logger.info("Orchestrator is using EKS as Cloud Controller")
        else:
            self._logger.error(f"Unsupported cloud {self._cast_ctrl.cluster['providerType']}")
            raise ValueError(f"{self._cast_ctrl.cluster['providerType']}")



CONFIG_TEMPLATE = {
    "CAST": {
        "CAST_API_KEY": "",
        "DEFAULT_CLUSTER_ID": ""
    },
    "AWS": {
        "AWS_REGION": "",
        "AWS_DEFAULT_NODE_GROUP": "",
        "AWS_ACCESS_KEY": "",
        "AWS_ACCESS_SECRET_KEY": ""
    },
    "AZURE": {
        "CLIENT_ID": "",
        "CLIENT_SECRET": "",
        "TENANT_ID": "",
        "DEFAULT_NODE_GROUP": ""
    },
    "GCP": {},
    "GENERAL": {
        "DEMO_NODE_COUNT": 7,
        "DEMO_REPLICAS": 2,
        "NG_SCALING_TIMEOUT": 300

    },
    "KUBECTL": {
        "DEMO_OFF_CRONJOB": "hibernate-pause"
    }
}

CLI_INPUTS_TEMPLATE = {
    "demo": False,
    "demo_subcommand": "",
    "cluster_id": "",
    "help": False,
    "debug": False
}
